ImpactX
Loading...
Searching...
No Matches
ExactDrift.H
Go to the documentation of this file.
1/* Copyright 2022-2026 The Regents of the University of California, through Lawrence
2 * Berkeley National Laboratory (subject to receipt of any required
3 * approvals from the U.S. Dept. of Energy). All rights reserved.
4 *
5 * This file is part of ImpactX.
6 *
7 * Authors: Chad Mitchell, Axel Huebl
8 * License: BSD-3-Clause-LBNL
9 */
10#ifndef IMPACTX_EXACTDRIFT_H
11#define IMPACTX_EXACTDRIFT_H
12
14#include "mixin/alignment.H"
15#include "mixin/pipeaperture.H"
16#include "mixin/beamoptic.H"
17#include "mixin/thick.H"
19#include "mixin/spintransport.H"
20#include "mixin/named.H"
21#include "mixin/nofinalize.H"
22
23#include <AMReX_Extension.H>
24#include <AMReX_Math.H>
25#include <AMReX_REAL.H>
26#include <AMReX_SIMD.H>
27
28#include <cmath>
29#include <stdexcept>
30
31namespace impactx::elements
32{
34 : public mixin::Named,
35 public mixin::BeamOptic<ExactDrift>,
36 public mixin::LinearTransport<ExactDrift>,
37 public mixin::Thick,
38 public mixin::Alignment,
41 public mixin::NoFinalize,
42 public amrex::simd::Vectorized<amrex::simd::native_simd_size_particlereal>
43 {
44 static constexpr auto type = "ExactDrift";
46
74
76 void reverse () { Thick::reverse(); }
77
79 using BeamOptic::operator();
80
88 void compute_constants (RefPart const & refpart)
89 {
90 using namespace amrex::literals; // for _rt and _prt
92
93 Alignment::compute_constants(refpart);
94
95 // length of the current slice
97
98 // access reference particle values to find beta, 1/(beta*gamma)**2
99 m_ibeta = 1_prt / refpart.beta();
100 m_ibetagam2 = 1_prt / powi<2>(refpart.beta_gamma());
101 }
102
117 template<typename T_Real=amrex::ParticleReal, typename T_IdCpu=uint64_t>
120 T_Real & AMREX_RESTRICT x,
121 T_Real & AMREX_RESTRICT y,
122 T_Real & AMREX_RESTRICT t,
123 T_Real const & AMREX_RESTRICT px,
124 T_Real const & AMREX_RESTRICT py,
125 T_Real const & AMREX_RESTRICT pt,
126 [[maybe_unused]] T_IdCpu const & AMREX_RESTRICT idcpu,
127 [[maybe_unused]] RefPart const & AMREX_RESTRICT refpart
128 ) const
129 {
130 using namespace amrex::literals; // for _rt and _prt
131 using namespace std; // for cmath(float)
132 using amrex::Math::powi;
133
134 // initialize output values
135 T_Real const xout = x;
136 T_Real const yout = y;
137 T_Real const tout = t;
138
139 // initialize output values of momenta
140 // T_Real const pxout = px;
141 // T_Real const pyout = py;
142 // T_Real const ptout = pt;
143
144 // compute the radical in the denominator (= pz):
145 T_Real const inv_pzden = 1_prt / sqrt(
146 powi<2>(pt - m_ibeta) -
148 powi<2>(px) -
149 powi<2>(py)
150 );
151
152 // advance position and momentum (exact drift)
153 x = xout + m_slice_ds * px * inv_pzden;
154 // pxout = px;
155 y = yout + m_slice_ds * py * inv_pzden;
156 // pyout = py;
157 t = tout - m_slice_ds * (m_ibeta + (pt - m_ibeta) * inv_pzden);
158 // ptout = pt;
159
160 // assign updated momenta
161 // px = pxout;
162 // py = pyout;
163 // pt = ptout;
164 }
165
171 void operator() (RefPart & AMREX_RESTRICT refpart) const
172 {
173 using namespace amrex::literals; // for _rt and _prt
174 using amrex::Math::powi;
175
176 // assign input reference particle values
177 amrex::ParticleReal const x = refpart.x;
178 amrex::ParticleReal const px = refpart.px;
179 amrex::ParticleReal const y = refpart.y;
180 amrex::ParticleReal const py = refpart.py;
181 amrex::ParticleReal const z = refpart.z;
182 amrex::ParticleReal const pz = refpart.pz;
183 amrex::ParticleReal const t = refpart.t;
184 amrex::ParticleReal const pt = refpart.pt;
185 amrex::ParticleReal const s = refpart.s;
186
187 // length of the current slice
188 amrex::ParticleReal const slice_ds = m_ds / nslice();
189
190 // assign intermediate parameter
191 amrex::ParticleReal const step = slice_ds /std::sqrt(powi<2>(pt)-1.0_prt);
192
193 // advance position and momentum (drift)
194 refpart.x = x + step*px;
195 refpart.y = y + step*py;
196 refpart.z = z + step*pz;
197 refpart.t = t - step*pt;
198
199 // advance integrated path length
200 refpart.s = s + slice_ds;
201 }
202
217 template<typename T_Real=amrex::ParticleReal, typename T_IdCpu=uint64_t>
220 T_Real & AMREX_RESTRICT x,
221 T_Real & AMREX_RESTRICT y,
222 T_Real & AMREX_RESTRICT t,
223 T_Real const & AMREX_RESTRICT px,
224 T_Real const & AMREX_RESTRICT py,
225 T_Real const & AMREX_RESTRICT pt,
226 [[maybe_unused]] T_Real const & AMREX_RESTRICT sx,
227 [[maybe_unused]] T_Real const & AMREX_RESTRICT sy,
228 [[maybe_unused]] T_Real const & AMREX_RESTRICT sz,
229 T_IdCpu const & AMREX_RESTRICT idcpu,
230 RefPart const & AMREX_RESTRICT refpart
231 ) const
232 {
233 // spin is unaffected
234
235 // phase space push
236 (*this)(x, y, t, px, py, pt, idcpu, refpart);
237 }
238
240 using LinearTransport::operator();
241
248 Map6x6
249 transport_map (RefPart const & AMREX_RESTRICT refpart) const
250 {
251 using namespace amrex::literals; // for _rt and _prt
252 using amrex::Math::powi;
253
254 // length of the current slice
255 amrex::ParticleReal const slice_ds = m_ds / nslice();
256
257 // access reference particle values to find beta*gamma^2
258 amrex::ParticleReal const pt_ref = refpart.pt;
259 amrex::ParticleReal const betgam2 = powi<2>(pt_ref) - 1_prt;
260
261 // assign linear map matrix elements
263 R(1,2) = slice_ds;
264 R(3,4) = slice_ds;
265 R(5,6) = slice_ds / betgam2;
266
267 // apply the transverse rotation (roll) alignment error
268 return rotate_aligned_map(R);
269 }
270
271 private:
272 // constants that are independent of the individually tracked particle,
273 // see: compute_constants() to refresh
276 };
277
278} // namespace impactx
279
281
282#endif // IMPACTX_EXACTDRIFT_H
#define AMREX_FORCE_INLINE
#define AMREX_RESTRICT
#define AMREX_GPU_HOST_DEVICE
#define AMREX_GPU_HOST
#define IMPACTX_PUSH_EXTERN_TEMPLATE(ElementType)
Definition PushAll.H:78
amrex_particle_real ParticleReal
constexpr T powi(T x) noexcept
__host__ __device__ GpuComplex< T > sqrt(const GpuComplex< T > &a_z) noexcept
Definition All.H:55
@ s
fixed s as the independent variable
Definition ImpactXParticleContainer.H:37
@ t
fixed t as the independent variable
Definition ImpactXParticleContainer.H:38
amrex::SmallMatrix< amrex::ParticleReal, 6, 6, amrex::Order::F, 1 > Map6x6
Definition CovarianceMatrix.H:20
static constexpr __host__ __device__ SmallMatrix< T, NRows, NCols, ORDER, StartIndex > Identity() noexcept
Definition ReferenceParticle.H:33
AMREX_GPU_HOST_DEVICE AMREX_FORCE_INLINE amrex::ParticleReal beta_gamma() const
Definition ReferenceParticle.H:167
AMREX_GPU_HOST_DEVICE AMREX_FORCE_INLINE amrex::ParticleReal beta() const
Definition ReferenceParticle.H:151
Definition ExactDrift.H:43
AMREX_GPU_HOST_DEVICE AMREX_FORCE_INLINE void spin_and_phasespace_push(T_Real &AMREX_RESTRICT x, T_Real &AMREX_RESTRICT y, T_Real &AMREX_RESTRICT t, T_Real const &AMREX_RESTRICT px, T_Real const &AMREX_RESTRICT py, T_Real const &AMREX_RESTRICT pt, T_Real const &AMREX_RESTRICT sx, T_Real const &AMREX_RESTRICT sy, T_Real const &AMREX_RESTRICT sz, T_IdCpu const &AMREX_RESTRICT idcpu, RefPart const &AMREX_RESTRICT refpart) const
Definition ExactDrift.H:219
AMREX_GPU_HOST_DEVICE AMREX_FORCE_INLINE void operator()(T_Real &AMREX_RESTRICT x, T_Real &AMREX_RESTRICT y, T_Real &AMREX_RESTRICT t, T_Real const &AMREX_RESTRICT px, T_Real const &AMREX_RESTRICT py, T_Real const &AMREX_RESTRICT pt, T_IdCpu const &AMREX_RESTRICT idcpu, RefPart const &AMREX_RESTRICT refpart) const
Definition ExactDrift.H:119
ExactDrift(amrex::ParticleReal ds, amrex::ParticleReal dx=DEFAULT_dx, amrex::ParticleReal dy=DEFAULT_dy, amrex::ParticleReal rotation_degree=DEFAULT_rotation_degree, amrex::ParticleReal aperture_x=DEFAULT_aperture_x, amrex::ParticleReal aperture_y=DEFAULT_aperture_y, int nslice=DEFAULT_nslice, std::optional< std::string > name=DEFAULT_name)
Definition ExactDrift.H:58
static constexpr auto type
Definition ExactDrift.H:44
AMREX_GPU_HOST AMREX_FORCE_INLINE Map6x6 transport_map(RefPart const &AMREX_RESTRICT refpart) const
Definition ExactDrift.H:249
amrex::ParticleReal m_slice_ds
Definition ExactDrift.H:274
ImpactXParticleContainer::ParticleType PType
Definition ExactDrift.H:45
void compute_constants(RefPart const &refpart)
Definition ExactDrift.H:88
amrex::ParticleReal m_ibetagam2
Definition ExactDrift.H:275
amrex::ParticleReal m_ibeta
m_ds / nslice();
Definition ExactDrift.H:275
void reverse()
Definition ExactDrift.H:76
Definition alignment.H:29
AMREX_GPU_HOST_DEVICE AMREX_FORCE_INLINE amrex::ParticleReal dy() const
Definition alignment.H:193
static constexpr amrex::ParticleReal DEFAULT_dy
Definition alignment.H:34
static constexpr amrex::ParticleReal DEFAULT_dx
Definition alignment.H:33
AMREX_GPU_HOST_DEVICE AMREX_FORCE_INLINE amrex::ParticleReal dx() const
Definition alignment.H:183
AMREX_GPU_HOST AMREX_FORCE_INLINE Map6x6 rotate_aligned_map(Map6x6 const &R) const
Definition alignment.H:267
Alignment(amrex::ParticleReal dx, amrex::ParticleReal dy, amrex::ParticleReal rotation_degree)
Definition alignment.H:43
static constexpr amrex::ParticleReal DEFAULT_rotation_degree
Definition alignment.H:35
Definition beamoptic.H:567
Definition lineartransport.H:50
Definition named.H:29
static constexpr std::nullopt_t DEFAULT_name
Definition named.H:30
AMREX_GPU_HOST Named(std::optional< std::string > name)
Definition named.H:59
AMREX_FORCE_INLINE std::string name() const
Definition named.H:124
Definition nofinalize.H:22
Definition pipeaperture.H:26
static constexpr amrex::ParticleReal DEFAULT_aperture_x
Definition pipeaperture.H:27
AMREX_GPU_HOST_DEVICE AMREX_FORCE_INLINE amrex::ParticleReal aperture_x() const
Definition pipeaperture.H:93
AMREX_GPU_HOST_DEVICE AMREX_FORCE_INLINE amrex::ParticleReal aperture_y() const
Definition pipeaperture.H:104
static constexpr amrex::ParticleReal DEFAULT_aperture_y
Definition pipeaperture.H:28
PipeAperture(amrex::ParticleReal aperture_x, amrex::ParticleReal aperture_y)
Definition pipeaperture.H:35
Definition spintransport.H:36
Definition thick.H:24
static constexpr int DEFAULT_nslice
Definition thick.H:25
Thick(amrex::ParticleReal ds, int nslice)
Definition thick.H:32
amrex::ParticleReal m_ds
Definition thick.H:70
AMREX_GPU_HOST_DEVICE AMREX_FORCE_INLINE amrex::ParticleReal ds() const
Definition thick.H:55
AMREX_GPU_HOST_DEVICE AMREX_FORCE_INLINE int nslice() const
Definition thick.H:45